@@ -9,70 +9,10 @@ Created on March 31, 2021
 import warnings
 
 import numpy as np
-from scipy.ndimage import grey_opening
-from scipy.sparse import diags, identity
+from scipy.interpolate import splev
+from scipy.sparse import csr_matrix
 
-from .utils import ParameterWarning, pad_edges, relative_difference
-
-
-def difference_matrix(data_size, diff_order=2, diff_format=None):
-    """
-    Creates an n-order finite-difference matrix.
-
-    Parameters
-    ----------
-    data_size : int
-        The number of data points.
-    diff_order : int, optional
-        The integer differential order; must be >= 0. Default is 2.
-    diff_format : str or None, optional
-        The sparse format to use for the difference matrix. Default is None,
-        which will use the default specified in :func:`scipy.sparse.diags`.
-
-    Returns
-    -------
-    diff_matrix : scipy.sparse.base.spmatrix
-        The sparse difference matrix.
-
-    Raises
-    ------
-    ValueError
-        Raised if `diff_order` or `data_size` is negative.
-
-    Notes
-    -----
-    Most baseline algorithms use 2nd order differential matrices when
-    doing penalized least squared fitting or Whittaker-smoothing-based fitting.
-
-    The resulting matrices are transposes of the result of
-    np.diff(np.eye(data_size), diff_order). This implementation allows using
-    the differential matrices are they are written in various publications,
-    ie. D.T * D.
-
-    """
-    if diff_order < 0:
-        raise ValueError('the differential order must be >= 0')
-    elif data_size < 0:
-        raise ValueError('data size must be >= 0')
-    elif diff_order > data_size:
-        # do not issue warning or exception to maintain parity with np.diff
-        diff_order = data_size
-
-    if diff_order == 0:
-        # faster to directly create identity matrix
-        diff_matrix = identity(data_size, format=diff_format)
-    else:
-        diagonals = np.zeros(2 * diff_order + 1)
-        diagonals[diff_order] = 1
-        for _ in range(diff_order):
-            diagonals = diagonals[:-1] - diagonals[1:]
-
-        diff_matrix = diags(
-            diagonals, np.arange(diff_order + 1),
-            shape=(data_size - diff_order, data_size), format=diff_format
-        )
-
-    return diff_matrix
+from .utils import ParameterWarning, difference_matrix, optimize_window, pad_edges
 
 
 def _yx_arrays(data, x_data=None, x_min=-1., x_max=1.):
@@ -434,72 +374,6 @@ def _setup_polynomial(data, x_data=None, weights=None, poly_order=2,
     return output
 
 
-def _optimize_window(data, increment=1, max_hits=3, window_tol=1e-6,
-                     max_half_window=None, min_half_window=None):
-    """
-    Optimizes the morphological half-window size.
-
-    Parameters
-    ----------
-    data : array-like, shape (N,)
-        The measured data values.
-    increment : int, optional
-        The step size for iterating half windows. Default is 1.
-    max_hits : int, optional
-        The number of consecutive half windows that must produce the same
-        morphological opening before accepting the half window as the optimum
-        value. Default is 3.
-    window_tol : float, optional
-        The tolerance value for considering two morphological openings as
-        equivalent. Default is 1e-6.
-    max_half_window : int, optional
-        The maximum allowable half-window size. If None (default), will be set
-        to (len(data) - 1) / 2.
-    min_half_window : int, optional
-        The minimum half-window size. If None (default), will be set to 1.
-
-    Returns
-    -------
-    half_window : int
-        The optimized half window size.
-
-    Notes
-    -----
-    May only provide good results for some morphological algorithms, so use with
-    caution.
-
-    References
-    ----------
-    Perez-Pueyo, R., et al. Morphology-Based Automated Baseline Removal for
-    Raman Spectra of Artistic Pigments. Applied Spectroscopy, 2010, 64, 595-600.
-
-    """
-    y = np.asarray(data)
-    if max_half_window is None:
-        max_half_window = (y.shape[0] - 1) // 2
-    if min_half_window is None:
-        min_half_window = 1
-
-    opening = grey_opening(y, [2 * min_half_window + 1])
-    hits = 0
-    best_half_window = min_half_window
-    for half_window in range(min_half_window + increment, max_half_window, increment):
-        new_opening = grey_opening(y, [half_window * 2 + 1])
-        if relative_difference(opening, new_opening) < window_tol:
-            if hits == 0:
-                # keep just the first window that fits tolerance
-                best_half_window = half_window - increment
-            hits += 1
-            if hits >= max_hits:
-                half_window = best_half_window
-                break
-        elif hits:
-            hits = 0
-        opening = new_opening
-
-    return max(half_window, 1)  # ensure half window is at least 1
-
-
 def _setup_morphology(data, half_window=None, **window_kwargs):
     """
     Sets the starting parameters for morphology-based methods.
@@ -551,21 +425,21 @@ def _setup_morphology(data, half_window=None, **window_kwargs):
     if half_window is not None:
         output_half_window = half_window
     else:
-        output_half_window = _optimize_window(y, **window_kwargs)
+        output_half_window = optimize_window(y, **window_kwargs)
 
     return y, output_half_window
 
 
-def _setup_window(data, half_window, **pad_kwargs):
+def _setup_smooth(data, half_window, **pad_kwargs):
     """
-    Sets the starting parameters for doing window-based algorithms.
+    Sets the starting parameters for doing smoothing-based algorithms.
 
     Parameters
     ----------
     data : array-like, shape (N,)
         The y-values of the measured data, with N data points.
     half_window : int, optional
-        The half-window used for the moving window functions. Used
+        The half-window used for the smoothing functions. Used
         to pad the left and right edges of the data to reduce edge
         effects.
     **pad_kwargs
@@ -626,3 +500,166 @@ def _setup_classification(data, x_data=None, weights=None):
         weight_array = np.ones(y.shape[0], bool)
 
     return y, x, weight_array, original_domain
+
+
+def spline_basis(x, num_knots=10, spline_degree=3, penalized=False):
+    """
+    Creates the basis matrix for B-splines and P-splines.
+
+    Parameters
+    ----------
+    x : numpy.ndarray, shape (N,)
+        The array of x-values
+    num_knots : int, optional
+        The number of interior knots for the spline. Default is 10.
+    spline_degree : int, optional
+        The degree of the spline. Default is 3, which is a cubic spline.
+    penalized : bool, optional
+        Whether the basis matrix should be for a penalized spline or a regular
+        B-spline. Default is False, which creates the basis for a B-spline.
+
+    Returns
+    -------
+    basis : numpy.ndarray
+        The basis matrix representing the spline, similar to the Vandermonde matrix
+        for polynomials.
+
+    Notes
+    -----
+    If `penalized` is True, makes the knots uniformly spaced to create p-splines. That
+    way, can use a finite difference matrix to impose penalties on the spline.
+
+    `degree` is used instead of `order` like for polynomials since the order of a spline
+    is defined by convention as `degree` + 1.
+
+    References
+    ----------
+    Eilers, P., et al. Twenty years of P-splines. SORT: Statistics and Operations Research
+    Transactions, 2015, 39(2), 149-186.
+
+    Hastie, T., et al. The Elements of Statistical Learning. Springer, 2017. Chapter 5.
+
+    """
+    spline_order = spline_degree + 1
+    if penalized:
+        x_min = x.min()
+        x_max = x.max()
+        dx = (x_max - x_min) / (num_knots + 1)  # +1 since number of sections is num_knots + 1
+        knots = np.linspace(
+            x_min - spline_degree * dx, x_max + spline_degree * dx,
+            num_knots + 2 * spline_order
+        )
+    else:
+        # TODO maybe provide a better way to select knot positions for regular B-splines
+        # use 0 and 100 percentile for the outer knots
+        inner_knots = np.percentile(x, np.linspace(0, 100, num_knots + 2))
+        knots = np.concatenate((
+            np.repeat(inner_knots[0], spline_degree), inner_knots,
+            np.repeat(inner_knots[-1], spline_degree)
+        ))
+
+    num_bases = knots.shape[0] - spline_order
+    basis = np.empty((num_bases, x.shape[0]))
+    coefs = np.zeros(num_bases)
+    # TODO would be faster to simply calculate the spline coefficients using de Boor's recursive
+    # algorithm; does it also give just the coefficients without all the extra zeros? would be
+    # nice if so, since it would use less space; also could then probably make a sparse or banded
+    # matrix rather than the full, dense array
+
+    # adapted from Scipy forums at: http://scipy-user.10969.n7.nabble.com/B-spline-basis-functions
+    for i in range(num_bases):
+        coefs[i] = 1  # evaluate the i-th basis within splev
+        basis[i] = splev(x, (knots, coefs, spline_degree))
+        coefs[i] = 0  # reset back to zero
+
+    # transpose to get a shape similar to the Vandermonde matrix
+    # TODO maybe it's preferable to not transpose, since typically basis.T is used in
+    # calculations more than basis; maybe make it a boolean input
+    return basis.T
+
+
+def _setup_splines(data, x_data=None, weights=None, spline_degree=3, num_knots=10,
+                   penalized=True, diff_order=3, lam=1, sparse_basis=True):
+    """
+    Sets the starting parameters for doing spline fitting.
+
+    Parameters
+    ----------
+    data : array-like, shape (N,)
+        The y-values of the measured data, with N data points.
+    x_data : array-like, shape (N,), optional
+        The x-values of the measured data. Default is None, which will create an
+        array from -1 to 1 with N points.
+    weights : array-like, shape (N,), optional
+        The weighting array. If None (default), then will be an array with
+        size equal to N and all values set to 1.
+    spline_degree : int, optional
+        The degree of the spline. Default is 3, which is a cubic spline.
+    num_knots : int, optional
+        The number of interior knots for the splines. Default is 10.
+    penalized : bool, optional
+        Whether the basis matrix should be for a penalized spline or a regular
+        B-spline. Default is True, which creates the basis for a penalized spline.
+    diff_order : int, optional
+        The integer differential order for the spline penalty; must be greater than 0.
+        Default is 3. Only used if `penalized` is True.
+    lam : float, optional
+        The smoothing parameter, lambda. Typical values are between 10 and
+        1e8, but it strongly depends on the number of knots and the difference order.
+        Default is 1.
+    sparse_basis : bool, optional
+        If True (default), will convert the spline basis to a sparse matrix with CSR
+        format.
+
+    Returns
+    -------
+    y : numpy.ndarray, shape (N,)
+        The y-values of the measured data, converted to a numpy array.
+    x : numpy.ndarray, shape (N,)
+        The x-values for fitting the spline.
+    basis : numpy.ndarray or scipy.sparse.csr.csr_matrix
+        The spline basis matrix. Is sparse with CSR format if `sparse_basis` is True.
+    weight_array : numpy.ndarray, shape (N,)
+        The weight array for fitting a polynomial to the data.
+    penalty_matrix : scipy.sparse.csr.csr_matrix
+        The penalty matrix for the spline. Only returned if `penalized` is True.
+
+    Raises
+    ------
+    ValueError
+        Raised is `diff_order` is less than 1 or if `weights` and `data` do not have
+        the same shape.
+
+    Warns
+    -----
+    UserWarning
+        Raised if `diff_order` is greater than 4.
+
+    """
+    y, x = _yx_arrays(data, x_data)
+    if weights is not None:
+        weight_array = np.asarray(weights)
+        if weight_array.shape != y.shape:
+            raise ValueError('weights must have the same shape as the input data')
+    else:
+        weight_array = np.ones(y.shape[0])
+    basis = spline_basis(x, num_knots, spline_degree, penalized)
+    if sparse_basis:
+        basis = csr_matrix(basis)
+    if not penalized:
+        return y, x, basis, weight_array
+
+    if diff_order < 1:
+        raise ValueError(
+            'the differential order must be > 0 for spline methods'
+        )
+    elif diff_order > 4:
+        warnings.warn(
+            ('differential orders greater than 4 can have numerical issues;'
+             ' consider using a differential order of 2 or 3 instead'),
+            ParameterWarning
+        )
+    diff_matrix = difference_matrix(basis.shape[1], diff_order, 'csc')
+    penalty_matrix = lam * diff_matrix.T * diff_matrix
+
+    return y, x, basis, weight_array, penalty_matrix
