@@ -264,10 +264,13 @@ def optimize_extended_range(data, x_data=None, method='asls', side='both', width
         known_background = np.hstack((known_background, added_left))
         lower_bound += added_window
 
-    if method in ('iasls', 'modpoly', 'imodpoly', 'poly', 'penalized_poly', 'loess', 'quant_reg'):
+    if method in (
+        'iasls', 'modpoly', 'imodpoly', 'poly', 'penalized_poly', 'loess', 'quant_reg',
+        'goldindec'
+    ):
         method_kwargs['x_data'] = fit_x_data
 
-    if 'poly' in method or method in ('loess', 'quant_reg'):
+    if 'poly' in method or method in ('loess', 'quant_reg', 'goldindec'):
         if any(not isinstance(val, int) for val in (min_value, max_value, step)):
             raise TypeError((
                 'min_value, max_value, and step must all be integers when'
@@ -342,7 +345,7 @@ def _determine_polyorders(y, x, poly_order, weights, fit_function, **fit_kwargs)
     """
     baseline = fit_function(y, x, poly_order, weights=weights, **fit_kwargs)[0]
     signal = y - baseline
-    basline_to_signal = (max(baseline) - min(baseline)) / (max(signal) - min(signal))
+    basline_to_signal = (baseline.max() - baseline.min()) / (signal.max() - signal.min())
     # Table 2 in reference
     if basline_to_signal < 0.2:
         orders = (1, 2)
