@@ -81,8 +81,8 @@ import numpy as np
 from ._algorithm_setup import _get_vander, _setup_polynomial
 from ._compat import jit, prange
 from .utils import (
-    _MIN_FLOAT, ParameterWarning, _convert_coef, _interp_inplace, relative_difference,
-    _quantile_irls
+    _MIN_FLOAT, ParameterWarning, _convert_coef, _interp_inplace, _quantile_irls,
+    relative_difference
 )
 
 
@@ -636,7 +636,7 @@ def penalized_poly(data, x_data=None, poly_order=2, tol=1e-3, max_iter=250,
            Correction. Applied Spectroscopy, 2015, 69(7), 834-842.
 
     """
-    if alpha_factor < 0 or alpha_factor > 1:
+    if not 0 < alpha_factor <= 1:
         raise ValueError('alpha_factor must be between 0 and 1')
     symmetric_loss, method = _identify_loss_method(cost_function)
     loss_function = {
@@ -1426,3 +1426,197 @@ def quant_reg(data, x_data=None, poly_order=2, quantile=0.05, tol=1e-6, max_iter
         params['coef'] = _convert_coef(coef, original_domain)
 
     return baseline, params
+
+
+def goldindec(data, x_data=None, poly_order=2, tol=1e-3, max_iter=250, weights=None,
+              cost_function='asymmetric_indec', peak_ratio=0.5, alpha_factor=0.99,
+              tol_2=1e-3, tol_3=1e-6, max_iter_2=100, return_coef=False):
+    """
+    Fits a polynomial baseline using a non-quadratic cost function.
+
+    The non-quadratic cost functions penalize residuals with larger values,
+    giving a more robust fit compared to normal least-squares.
+
+    Parameters
+    ----------
+    data : array-like, shape (N,)
+        The y-values of the measured data, with N data points.
+    x_data : array-like, shape (N,), optional
+        The x-values of the measured data. Default is None, which will create an
+        array from -1 to 1 with N points.
+    poly_order : int, optional
+        The polynomial order for fitting the baseline. Default is 2.
+    tol : float, optional
+        The exit criteria for the fitting with a given threshold value. Default is 1e-3.
+    max_iter : int, optional
+        The maximum number of iterations for fitting a threshold value. Default is 250.
+    weights : array-like, shape (N,), optional
+        The weighting array. If None (default), then will be an array with
+        size equal to N and all values set to 1.
+    cost_function : str, optional
+        The non-quadratic cost function to minimize. Unlike :func:`.penalized_poly`,
+        this function only works with asymmetric cost functions, so the symmetry prefix
+        ('a' or 'asymmetric') is optional (eg. 'indec' and 'a_indec' are the same). Default
+        is 'asymmetric_indec'. Available methods, and their associated reference, are:
+
+            * 'asymmetric_indec'[18]_
+            * 'asymmetric_truncated_quadratic'[19]_
+            * 'asymmetric_huber'[19]_
+
+    peak_ratio : float, optional
+        A value between 0 and 1 that designates how many points in the data belong
+        to peaks. Values are valid within ~10% of the actual peak ratio. Default is 0.5.
+    alpha_factor : float, optional
+        A value between 0 and 1 that controls the value of the penalty. Default is
+        0.99. Typically should not need to change this value.
+    tol_2 : float, optional
+        The exit criteria for the difference between the optimal up-down ratio (number of
+        points above 0 in the residual compared to number of points below 0) and the up-down
+        ratio for a given threshold value. Default is 1e-3.
+    tol_3 : float, optional
+        The exit criteria for the relative change in the threshold value. Default is 1e-6.
+    max_iter_2 : float, optional
+        The number of iterations for iterating between different threshold values.
+        Default is 100.
+    return_coef : bool, optional
+        If True, will convert the polynomial coefficients for the fit baseline to
+        a form that fits the input x_data and return them in the params dictionary.
+        Default is False, since the conversion takes time.
+
+    Returns
+    -------
+    baseline : numpy.ndarray, shape (N,)
+        The calculated baseline.
+    params : dict
+        A dictionary with the following items:
+
+        * 'weights': numpy.ndarray, shape (N,)
+            The weight array used for fitting the data.
+        * 'tol_history': numpy.ndarray, shape (J, K)
+            An array containing the calculated tolerance values for each iteration
+            of both threshold values and fit values. Index 0 are the tolerence values
+            for the difference in up-down ratios, index 1 are the tolerance values for
+            the relative change in the threshold, and indices >= 2 are the tolerance values
+            for each fit. All values that were not used in fitting have values of 0. Shape J
+            is 2 plus the number of iterations for the threshold to converge (related to
+            `max_iter_2`, `tol_2`, `tol_3`), and shape K is the maximum of the number of
+            iterations for the threshold and the maximum number of iterations for all of
+            the fits of the various threshold values (related to `max_iter` and `tol`).
+        * 'threshold' : float
+            The optimal threshold value. Could be used in :func:`.penalized_poly`
+            for fitting other similar data.
+        * 'coef': numpy.ndarray, shape (poly_order + 1,)
+            Only if `return_coef` is True. The array of polynomial parameters
+            for the baseline, in increasing order. Can be used to create a
+            polynomial using numpy.polynomial.polynomial.Polynomial().
+
+    Raises
+    ------
+    ValueError
+        Raised if `alpha_factor` or `peak_ratio` are not between 0 and 1, or if the
+        specified cost function is symmetric.
+
+    References
+    ----------
+    .. [18] Liu, J., et al. Goldindec: A Novel Algorithm for Raman Spectrum Baseline
+            Correction. Applied Spectroscopy, 2015, 69(7), 834-842.
+    .. [19] Mazet, V., et al. Background removal from spectra by designing and
+            minimising a non-quadratic cost function. Chemometrics and Intelligent
+            Laboratory Systems, 2005, 76(2), 121–133.
+
+    """
+    if not 0 < alpha_factor <= 1:
+        raise ValueError('alpha_factor must be between 0 and 1')
+    elif not 0 < peak_ratio < 1:
+        raise ValueError('peak_ratio must be between 0 and 1')
+    try:
+        symmetric_loss, method = _identify_loss_method(cost_function)
+    except ValueError:  # do not require a prefix since cost must be asymmetric
+        symmetric_loss, method = _identify_loss_method('a_' + cost_function)
+    if symmetric_loss:
+        # symmetric cost functions don't work due to how the up-down ratio vs
+        # peak_ratio function was created in the reference; in theory, could simulate
+        # spectra with both positive and negative peaks following the reference
+        # and build another empirical function, but would likely need to also
+        # add other parameters detailing the percent of positive vs negative peaks,
+        # etc., so it's not worth the effort
+        raise ValueError('goldindec only works for asymmetric cost functions')
+
+    loss_function = {
+        'huber': _huber_loss,
+        'truncated_quadratic': _truncated_quadratic_loss,
+        'indec': _indec_loss
+    }[method]
+    y, x, weight_array, original_domain, vander, pseudo_inverse = _setup_polynomial(
+        data, x_data, weights, poly_order, return_vander=True, return_pinv=True
+    )
+    num_y = y.shape[0]
+    up_down_ratio_goal = (
+        0.7679 + 11.2358 * peak_ratio - 39.7064 * peak_ratio**2 + 92.3583 * peak_ratio**3
+    )
+    # TODO reference states threshold must be <= 2 for half-quadratic minimization to
+    # be valid for indec cost function, and normalized y so that threshold is always <= 2;
+    # however, it seems to work fine without normalization; just be aware in case errors
+    # occur, may have to normalize y in both this function and penalized_poly
+    sqrt_w = np.sqrt(weight_array)
+    y_fit = sqrt_w * y
+
+    coef = np.dot(pseudo_inverse, y_fit)
+    initial_baseline = np.dot(vander, coef)
+
+    a = 0
+    # reference used b=1, but normalized y before fitting; instead, set b as max of
+    # initial residual
+    b = abs((y - initial_baseline).max())
+    threshold = a + 0.618 * (b - a)
+    loss_kwargs = {
+        'threshold': threshold, 'alpha_factor': alpha_factor,
+        'symmetric': symmetric_loss
+    }
+    # have to use zeros rather than empty for tol_history since each inner fit may
+    # have a different number of iterations
+    tol_history = np.zeros((max_iter_2 + 2, max(max_iter, max_iter_2)))
+    j_max = 0
+    for i in range(max_iter_2):
+        baseline = initial_baseline
+        for j in range(max_iter):
+            baseline_old = baseline
+            coef = np.dot(
+                pseudo_inverse, y_fit + loss_function(y_fit - sqrt_w * baseline, **loss_kwargs)
+            )
+            baseline = np.dot(vander, coef)
+            calc_difference = relative_difference(baseline_old, baseline)
+            tol_history[i + 2, j] = calc_difference
+            if calc_difference < tol:
+                break
+        if j > j_max:
+            j_max = j
+
+        up_count = (y > baseline).sum()
+        up_down_ratio = up_count / max(1, num_y - up_count)
+        calc_difference = up_down_ratio - up_down_ratio_goal
+        tol_history[0, i] = calc_difference
+        if calc_difference > tol_2:
+            a = threshold
+        elif calc_difference < -tol_2:
+            b = threshold
+        else:
+            break
+        threshold = a + 0.618 * (b - a)
+        # this exit criteria was not stated in the reference, but the change in threshold
+        # becomes zero fairly quickly, so need to also exit rather than needlessly
+        # continuing to calculate with the same threshold value
+        calc_difference = relative_difference(loss_kwargs['threshold'], threshold)
+        tol_history[1, i] = calc_difference
+        if calc_difference < tol_3:
+            break
+        loss_kwargs['threshold'] = threshold
+
+    params = {
+        'weights': weight_array, 'tol_history': tol_history[:i + 3, :max(i, j_max) + 1],
+        'threshold': loss_kwargs['threshold']
+    }
+    if return_coef:
+        params['coef'] = _convert_coef(coef, original_domain)
+
+    return baseline, params
