@@ -14,7 +14,9 @@ from scipy.linalg import solve_banded, solveh_banded
 from . import utils
 from ._algorithm_setup import _diff_1_diags, _setup_whittaker, _yx_arrays
 from ._compat import _HAS_PENTAPY, _pentapy_solve
-from .utils import ParameterWarning, _safe_std, relative_difference
+from .utils import (
+    ParameterWarning, _mollifier_kernel, _safe_std, pad_edges, padded_convolve, relative_difference
+)
 
 
 def _shift_rows(matrix, diagonals=2):
@@ -814,3 +816,140 @@ def psalsa(data, lam=1e5, p=0.5, k=None, diff_order=2, max_iter=50, tol=1e-3, we
     params = {'weights': weight_array, 'tol_history': tol_history[:i + 1]}
 
     return baseline, params
+
+
+def derpsalsa(data, lam=1e6, p=0.01, k=None, diff_order=2, max_iter=50, tol=1e-3,
+              weights=None, smooth_half_window=None, num_smooths=16, **pad_kwargs):
+    """
+    Derivative Peak-Screening Asymmetric Least Squares Algorithm (derpsalsa).
+
+    Parameters
+    ----------
+    data : array-like, shape (N,)
+        The y-values of the measured data, with N data points. Must not
+        contain missing data (NaN) or Inf.
+    lam : float, optional
+        The smoothing parameter. Larger values will create smoother baselines.
+        Default is 1e6.
+    p : float, optional
+        The penalizing weighting factor. Must be between 0 and 1. Values greater
+        than the baseline will be given p weight, and values less than the baseline
+        will be given p-1 weight. Default is 0.1.
+    k : float, optional
+        A factor that controls the exponential decay of the weights for baseline
+        values greater than the data. Should be approximately the height at which
+        a value could be considered a peak. Default is None, which sets `k` to
+        one-tenth of the standard deviation of the input data. A large k value
+        will produce similar results to :func:`.asls`.
+    diff_order : int, optional
+        The order of the differential matrix. Must be greater than 0. Default is 2
+        (second order differential matrix). Typical values are 2 or 1.
+    max_iter : int, optional
+        The max number of fit iterations. Default is 50.
+    tol : float, optional
+        The exit criteria. Default is 1e-3.
+    weights : array-like, shape (N,), optional
+        The weighting array. If None (default), then the initial weights
+        will be an array with size equal to N and all values set to 1.
+    smooth_half_window : int, optional
+        The half-window to use for smoothing the data beforebefore computing the first
+        and second derivatives. Default is None, which will use ``len(data) / 200``.
+    num_smooths : int, optional
+        The number of times to smooth the data before computing the first
+        and second derivatives. Default is 16.
+    **pad_kwargs
+        Additional keyword arguments to pass to :func:`.pad_edges` for padding
+        the edges of the data to prevent edge effects from smoothing.
+
+    Returns
+    -------
+    baseline : numpy.ndarray, shape (N,)
+        The calculated baseline.
+    params : dict
+        A dictionary with the following items:
+
+        * 'weights': numpy.ndarray, shape (N,)
+            The weight array used for fitting the data.
+        * 'tol_history': numpy.ndarray
+            An array containing the calculated tolerance values for
+            each iteration. The length of the array is the number of iterations
+            completed. If the last value in the array is greater than the input
+            `tol` value, then the function did not converge.
+
+    Raises
+    ------
+    ValueError
+        Raised if p is not between 0 and 1.
+
+    References
+    ----------
+    Korepanov, V. Asymmetric least-squares baseline algorithm with peak screening for
+    automatic processing of the Raman spectra. Journal of Raman Spectroscopy. 2020,
+    51(10), 2061-2065.
+
+    """
+    if not 0 < p < 1:
+        raise ValueError('p must be between 0 and 1')
+    using_pentapy = _HAS_PENTAPY and diff_order == 2
+    y, diagonals, weight_array = _setup_whittaker(
+        data, lam, diff_order, weights, False, not using_pentapy, using_pentapy
+    )
+    if k is None:
+        k = np.std(y) / 10
+    num_y = y.shape[0]
+    if smooth_half_window is None:
+        smooth_half_window = num_y // 200
+    # could pad the data every iteration, but it is ~2-3 times slower and only affects
+    # the edges, so it's not worth it
+    y_smooth = pad_edges(y, smooth_half_window, **pad_kwargs)
+    if smooth_half_window > 0:
+        smooth_kernel = _mollifier_kernel(smooth_half_window)
+        for _ in range(num_smooths):
+            y_smooth = padded_convolve(y_smooth, smooth_kernel)
+    y_smooth = y_smooth[smooth_half_window:num_y + smooth_half_window]
+
+    diff_y_1 = np.diff(np.concatenate((y_smooth[:1], y_smooth)))
+    diff_y_2 = np.diff(np.concatenate((diff_y_1[:1], diff_y_1)))  # TODO check this is correct
+    # x.dot(x) is same as (x**2).sum() but faster
+    rms_diff_1 = np.sqrt(diff_y_1.dot(diff_y_1) / num_y)
+    rms_diff_2 = np.sqrt(diff_y_2.dot(diff_y_2) / num_y)
+
+    diff_1_weights = np.exp(-((diff_y_1 / rms_diff_1)**2) / 2)
+    diff_2_weights = np.exp(-((diff_y_2 / rms_diff_2)**2) / 2)
+    partial_weights = diff_1_weights * diff_2_weights
+
+    main_diag_idx = diff_order if using_pentapy else -1
+    main_diagonal = diagonals[main_diag_idx].copy()
+    tol_history = np.empty(max_iter + 1)
+    for i in range(max_iter + 1):
+        diagonals[main_diag_idx] = main_diagonal + weight_array
+        if using_pentapy:
+            baseline = _pentapy_solve(
+                diagonals, weight_array * y, True, True, utils.PENTAPY_SOLVER
+            )
+        else:
+            baseline = solveh_banded(
+                diagonals, weight_array * y, overwrite_b=True, check_finite=False
+            )
+        residual = y - baseline
+        # no need for caution since inner exponential is always negative, but still mask
+        # since it's faster than performing the square and exp on the full residual
+        new_weights = np.full(num_y, 1 - p, dtype=float)
+        mask = residual > 0
+        new_weights[mask] = p * np.exp(-((residual[mask] / k)**2) / 2)
+        # reference is not clear as to how p and 1-p are applied; an alternative could
+        # be that partial_weights are multiplied only where residual > 0 and that all
+        # other weights are 1-p, but based on Figure 1c in the reference, total weights
+        # are never greater than partial_weights, so that must mean the non-peak regions
+        # have a weight of (1-p) * partial_weights rather than just 1-p; both weighting
+        # systems give near identical results, so it is not a big deal
+        new_weights *= partial_weights
+        calc_difference = relative_difference(weight_array, new_weights)
+        tol_history[i] = calc_difference
+        if calc_difference < tol:
+            break
+        weight_array = new_weights
+
+    params = {'weights': weight_array, 'tol_history': tol_history[:i + 1]}
+
+    return baseline, params
